<?php

namespace Drupal\mustache\Render;

use Drupal\Component\Render\MarkupInterface;

/**
 * Passes multiple safe strings through the render system.
 *
 * This iterable markup allows for both concatenating the contained items,
 * but also define a value that may be used instead of concatenating the items.
 * For nesting iterable markup items, a child markup item may reference to its
 * parent item through its "parent" property. All listed properties below are
 * directly accessible from outside and can be manipulated.
 *
 * This object should only be constructed with known safe strings. If there is
 * any risk that the string contains user-entered data that has not been
 * filtered first, it must not be used.
 *
 * @internal
 *   This class must only be used inside the mustache_templates project.
 */
final class IterableMarkup implements MarkupInterface, \Countable, \IteratorAggregate, \ArrayAccess {

  /**
   * The iterable items.
   *
   * @var array
   */
  private $items;

  /**
   * The string value to use instead of concatenating the items.
   *
   * @var mixed
   */
  private $value;

  /**
   * The parent.
   *
   * @var mixed
   */
  private $parent;

  /**
   * The separator to use when concatenating as string.
   *
   * @var string
   */
  private $separator = ', ';

  /**
   * The position of this item within its parent's list of items.
   *
   * Position counting starts at value 1.
   *
   * @var int
   */
  private $position = 1;

  /**
   * Whether this item is the first one within its parent's list of items.
   *
   * @var bool
   */
  private $first = TRUE;

  /**
   * Whether this item is the last one within its parent's list of items.
   *
   * @var bool
   */
  private $last = TRUE;

  /**
   * Create an iterable markup object.
   *
   * @param array $items
   *   The items to iterate on.
   * @param mixed $value
   *   (Optional) The string value to use instead of concatenating the items.
   * @param mixed $parent
   *   (Optional) The parent, usually another IterableMarkup instance.
   * @param string $separator
   *   (Optional) The separator to use on string concatenation.
   *
   * @return static
   *   The iterable markup object.
   */
  public static function create(array $items = [], $value = NULL, $parent = NULL, $separator = ', ') {
    $instance = new static();
    $instance->items = $items;
    $instance->value = $value;
    $instance->parent = $parent;
    $instance->separator = $separator;
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function __toString() {
    if ($this->value !== NULL) {
      return (string) $this->value;
    }
    return implode($this->separator, $this->items);
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    if ($this->empty()) {
      return [];
    }
    $values = [
      'value' => $this->__toString(),
      'items' => [],
      'separator' => $this->separator,
      'position' => $this->position,
      'first' => $this->first,
      'last' => $this->last,
    ];
    foreach ($this->items as $i_key => $i_val) {
      if ($i_val instanceof static) {
        $i_val = $i_val->jsonSerialize();
      }
      elseif (is_object($i_val)) {
        if (method_exists($i_val, '__toString')) {
          $i_val = (string) $i_val;
        }
        elseif (method_exists($i_val, 'getString')) {
          $i_val = $i_val->getString();
        }
      }
      $values['items'][$i_key] = $i_val;
    }
    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function count() {
    if ($this->items) {
      return count($this->items);
    }
    if ($this->value !== NULL) {
      return strlen((string) $this->value);
    }
    return 0;
  }

  /**
   * Whether the whole markup object is empty.
   *
   * @return bool
   *   Returns TRUE when it is empty.
   */
  public function empty() {
    return !$this->count();
  }

  /**
   * Whether the whole markup object contains at least one value.
   *
   * @return bool
   *   Returns TRUE when it exists.
   */
  public function exists() {
    return !$this->empty();
  }

  /**
   * {@inheritdoc}
   */
  public function getIterator() {
    if ($this->items) {
      return new \CallbackFilterIterator(new \ArrayIterator($this->items), function ($current, $key, $iterator) {
        if ($current instanceof static) {
          return $current->exists();
        }
        return ((is_scalar($current) && !empty($current)) || (is_object($current) && method_exists($current, '__toString') && !empty((string) $current)));
      });
    }
    return $this->exists() ? new \ArrayIterator([$this]) : new \ArrayIterator($this->items);
  }

  /**
   * {@inheritdoc}
   */
  public function offsetSet($offset, $value) {
    if (is_null($offset)) {
      $this->items[] = $value;
    }
    else {
      $this->items[$offset] = $value;
    }
    $this->updatePositions();
  }

  /**
   * {@inheritdoc}
   */
  public function offsetExists($offset) {
    if (!isset($this->items[$offset])) {
      if (!empty($this->items)) {
        foreach ($this->items as &$item) {
          if (((($item instanceof \ArrayAccess) || is_array($item)) && isset($item[$offset])) || (is_object($item) && isset($item->$offset))) {
            return TRUE;
          }
        }
      }
      return FALSE;
    }
    if ($this->items[$offset] instanceof static) {
      return $this->items[$offset]->exists();
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function offsetUnset($offset) {
    unset($this->items[$offset]);
    $this->updatePositions();
  }

  /**
   * {@inheritdoc}
   */
  public function &offsetGet($offset) {
    if (isset($this->items[$offset])) {
      return $this->items[$offset];
    }
    if (!empty($this->items)) {
      $selection = static::create([], NULL, $this);
      foreach ($this->items as &$item) {
        if ((($item instanceof \ArrayAccess)|| is_array($item)) && isset($item[$offset])) {
          $selection[] = $item[$offset];
        }
        elseif (is_object($item) && isset($item->$offset)) {
          $selection[] = $item->$offset;
        }
      }
      return $selection;
    }
    $null = static::create([], NULL, $this);
    return $null;
  }

  /**
   * Implements the magic __isset() method.
   *
   * @return bool
   *   Whether the property is set.
   */
  public function __isset($name) {
    if (isset($this->items[$name]) && $this->items[$name] instanceof static) {
      return $this->items[$name]->exists();
    }
    if (isset($this->items[$name]) || (property_exists($this, $name) && !is_null($this->$name))) {
      return TRUE;
    }
    if ($name === 'value') {
      return isset($this->value) || !empty($this->items);
    }
    if ($name === 'string') {
      return isset($this->value);
    }
    if (!empty($this->items)) {
      foreach ($this->items as &$item) {
        if (((($item instanceof \ArrayAccess) || is_array($item)) && isset($item[$name])) || (is_object($item) && isset($item->$name))) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * Implements the magic __set() method.
   */
  public function __set($name, $value) {
    $this->$name = &$value;
  }

  /**
   * Implements the magic __get() method.
   *
   * @return mixed
   *   The value for the given name.
   */
  public function &__get($name) {
    if (isset($this->items[$name])) {
      return $this->items[$name];
    }
    if ($name === 'value') {
      if (empty($this->items)) {
        $value = &$this->value;
        return $value;
      }
      return $this;
    }
    if ($name === 'string') {
      $return = (string) $this->value;
      return $return;
    }
    if (property_exists($this, $name)) {
      return $this->$name;
    }
    if (!empty($this->items)) {
      $selection = static::create([], NULL, $this);
      foreach ($this->items as &$item) {
        if ((($item instanceof \ArrayAccess) || is_array($item)) && isset($item[$name])) {
          $selection[] = $item[$name];
        }
        elseif (is_object($item) && isset($item->$name)) {
          $selection[] = $item->$name;
        }
      }
      return $selection;
    }
    $return = static::create([], NULL, $this);
    return $return;
  }

  /**
   * Updates the positions on the child items.
   */
  public function updatePositions() {
    reset($this->items);
    $first = TRUE;
    $i = 0;
    foreach ($this->items as &$item) {
      $i++;
      if ($item instanceof static) {
        $item->position = $i;
        $item->first = $first;
        $item->last = FALSE;
        $item->parent = $this;
        $item->updatePositions();
      }
      elseif (is_array($item)) {
        $item['position'] = $i;
        $item['first'] = $first;
        $item['last'] = FALSE;
        $item['parent'] = $this;
      }
      $first = FALSE;
    }
    if (isset($item)) {
      if ($item instanceof static) {
        $item->last = TRUE;
      }
      elseif (is_array($item)) {
        $item['last'] = TRUE;
      }
    }
  }

}
